'''
python3.8 -m venv venv
source venv/bin/activate
pip3 install -r requirements.txt
python3 train_abp_celeba.py
'''
import os
import sys
import logging
import datetime
import shutil
import argparse
import math
import pickle

import torch
import torch.nn as nn
import torchvision as tv
import torch.nn.functional as F


#################################################################################################

def create_args():
    args = argparse.Namespace()

    args.seed = 1
    args.gpu = 0
    args.img_sz = 32
    args.img_ch = 3
    args.n_batch = 64
    args.n_iter = int(1e5)
    args.lr = 1e-4
    args.z_persistent = False
    args.z_sz = 100
    args.z_sigma = 0.5
    args.z_step_size = 0.1
    args.z_n_iters = 20
    args.z_with_noise = True

    return args

#################################################################################################

def M(g, z, x, args):
    mse = nn.MSELoss(reduction='sum')
    for i in range(args.z_n_iters):
        z = torch.autograd.Variable(z, requires_grad=True)
        x_hat = g(z)
        L = 1.0 / (2.0 * args.z_sigma * args.z_sigma) * mse(x_hat, x)
        L.backward()
        z = z - 0.5 * args.z_step_size * args.z_step_size * (z + z.grad)
        if args.z_with_noise:
            eps = torch.randn(args.n_batch, args.z_sz, 1, 1).to(x.device)
            z += args.z_step_size * eps

    return z

#################################################################################################

class G(nn.Module):
    def __init__(self, nz=100, nc=3, ngf=16):
        super(G, self).__init__()
        self.deconv1 = nn.ConvTranspose2d(nz, ngf*16, 4, 1, 0, bias=False)
        self.deconv1_bn = nn.BatchNorm2d(ngf*16)

        self.deconv2 = nn.ConvTranspose2d(ngf*16, ngf*16, 4, 2, 1, bias=False)
        self.deconv2_bn = nn.BatchNorm2d(ngf*16)

        self.deconv3 = nn.ConvTranspose2d(ngf*16, ngf*8, 4, 2, 1, bias=False)
        self.deconv3_bn = nn.BatchNorm2d(ngf*8)

        self.deconv4 = nn.ConvTranspose2d(ngf*8, ngf*4, 4, 2, 1, bias=False)
        self.deconv4_bn = nn.BatchNorm2d(ngf*4)

        self.deconv5 = nn.ConvTranspose2d(ngf*4, nc, 3, 1, 1)

    def forward(self, x):
        x = F.relu(self.deconv1_bn(self.deconv1(x)))
        x = F.relu(self.deconv2_bn(self.deconv2(x)))
        x = F.relu(self.deconv3_bn(self.deconv3(x)))
        x = F.relu(self.deconv4_bn(self.deconv4(x)))
        x = torch.tanh(self.deconv5(x))
        return x

#################################################################################################

def train(args, output_dir, logger):

    torch.manual_seed(args.seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(args.seed)

    device = torch.device(f'cuda:{args.gpu}' if torch.cuda.is_available() else 'cpu')

    path = 'data/celeba/celeba_40000_32.pickle'
    with open(path, mode='rb') as f:
        ds = pickle.load(f)
    p_d = torch.tensor(torch.stack(ds), dtype=torch.float32).to(device)

    p_0 = torch.distributions.MultivariateNormal(torch.zeros(args.z_sz).to(device), torch.eye(args.z_sz).to(device))
    if args.z_persistent:
        p_0_chains = p_0.sample([p_d.shape[0]]).view([p_d.shape[0], -1, 1, 1]).to(device)

    def sample_p_0():
        return p_0.sample([args.n_batch]).view([args.n_batch, -1, 1, 1]).to(device)

    def sample_p_d_and_p_0():
        i = torch.LongTensor(args.n_batch).random_(0, p_d.shape[0])
        p_d_x = p_d[i].detach()
        if args.z_persistent:
            p_0_z = p_0_chains[i].detach()
        else:
            p_0_z = p_0.sample([args.n_batch]).view([args.n_batch, -1, 1, 1]).to(device)
        return p_d_x, p_0_z, i

    sqrt = lambda x: int(torch.sqrt(torch.Tensor([x])))
    plot = lambda p, x: tv.utils.save_image(torch.clamp(x, -1., 1.), p, normalize=True, nrow=sqrt(args.n_batch))

    g = G().to(device)
    g.train()

    optim = torch.optim.Adam(g.parameters(), lr=args.lr, betas=[.9, .999])

    mse = nn.MSELoss(reduction='sum')

    for i in range(args.n_iter):
        x, z_0, chains_i = sample_p_d_and_p_0()
        z = M(g, z_0, x, args)
        x_hat = g(z)
        L = 0.5 * args.z_sigma * args.z_sigma * mse(x, x_hat)

        if args.z_persistent:
            p_0_chains[chains_i] = z.detach()

        optim.zero_grad()
        L.backward()
        optim.step()

        if i % 10 == 0:
            norm_z0 = torch.norm(z_0, dim=0).mean()
            norm_z = torch.norm(z, dim=0).mean()
            logger.info('{:>6d} mse(x, x_hat)={:>14.9f} norm(z0)={:>14.9f} norm(z)={:>14.9f}'.format(i, mse(x, x_hat.detach()) / args.n_batch, norm_z0, norm_z))

        if i % 100 == 0:
            plot('{}/{:>06d}_x.png'.format(output_dir, i), x)
            plot('{}/{:>06d}_x_hat.png'.format(output_dir, i), x_hat)
            plot('{}/{:>06d}_x_sample.png'.format(output_dir, i), g(sample_p_0()))


#################################################################################################

def get_exp_id(file):
    return os.path.splitext(os.path.basename(file))[0]


def get_output_dir(exp_id):
    t = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
    output_dir = os.path.join('output/' + exp_id, t)
    os.makedirs(output_dir, exist_ok=True)
    return output_dir


def setup_logging(name, output_dir, console=True):
    log_format = logging.Formatter("%(asctime)s : %(message)s")
    logger = logging.getLogger(name)
    logger.handlers = []
    output_file = os.path.join(output_dir, 'output.log')
    file_handler = logging.FileHandler(output_file)
    file_handler.setFormatter(log_format)
    logger.addHandler(file_handler)
    if console:
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(log_format)
        logger.addHandler(console_handler)
    logger.setLevel(logging.INFO)
    return logger


def copy_source(file, output_dir):
    shutil.copyfile(file, os.path.join(output_dir, os.path.basename(file)))


if __name__ == '__main__':
    exp_id = get_exp_id(__file__)
    output_dir = get_output_dir(exp_id)
    copy_source(__file__, output_dir)
    logger = setup_logging('main', output_dir)

    train(create_args(), output_dir, logger)
